"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.$average = exports.$max = exports.$min = exports.$sum = exports.$last = exports.$first = exports.$includesvisual = exports.$includesequal = exports.$mapset = exports.$map = void 0;
const commons_1 = require("./commons");
const compare_1 = require("./compare");
const types_1 = require("./types");
function $map(values, options) {
    const opts = (0, commons_1.$isfunction)(options) ? { callback: options }
        : ((0, commons_1.$ok)(options) ? options : {});
    if (!(0, commons_1.$ok)(opts.callback)) {
        opts.callback = v => v;
    }
    if (!(0, commons_1.$ok)(opts.unicity)) {
        opts.unicity = types_1.TSUnicity.None;
    }
    const ret = new Array();
    if ((0, commons_1.$ok)(values)) {
        const fn = opts.callback;
        switch (opts.unicity) {
            case types_1.TSUnicity.None: {
                let index = 0;
                for (let v of values) {
                    const mv = fn(v, index);
                    if ((0, commons_1.$ok)(mv)) {
                        ret.push(mv);
                    }
                    index++;
                }
                break;
            }
            case types_1.TSUnicity.Objects:
                _mapEquality(ret, values, fn, (_) => false);
                break;
            case types_1.TSUnicity.Equality:
                _mapEquality(ret, values, fn, (src, o) => $includesequal(src, o));
                break;
            case types_1.TSUnicity.Visual:
                _mapEquality(ret, values, fn, (src, o) => $includesvisual(src, o));
                break;
        }
    }
    return ret;
}
exports.$map = $map;
function $mapset(values, callback) {
    if (!(0, commons_1.$ok)(callback)) {
        callback = function (v, _) { return v; };
    }
    const ret = new Set();
    if ((0, commons_1.$ok)(values)) {
        let index = 0;
        for (let v of values) {
            const mv = callback(v, index++);
            if ((0, commons_1.$ok)(mv)) {
                ret.add(mv);
            }
        }
    }
    return ret;
}
exports.$mapset = $mapset;
function $includesequal(values, object) {
    if ((0, commons_1.$ok)(values)) {
        for (let v of values) {
            if ((0, compare_1.$equal)(object, v))
                return true;
        }
    }
    return false;
}
exports.$includesequal = $includesequal;
function $includesvisual(values, object) {
    if ((0, commons_1.$ok)(values)) {
        for (let v of values) {
            if ((0, compare_1.$visualequal)(object, v))
                return true;
        }
    }
    return false;
}
exports.$includesvisual = $includesvisual;
function $first(values) {
    const n = (0, commons_1.$count)(values);
    return n > 0 ? values[0] : undefined;
}
exports.$first = $first;
function $last(values) {
    const n = (0, commons_1.$count)(values);
    return n > 0 ? values[n - 1] : undefined;
}
exports.$last = $last;
function $sum(values) {
    const [, , , sum] = _countsAndSum(values);
    return sum;
}
exports.$sum = $sum;
function $min(values) { return _minmax(values, types_1.Descending); }
exports.$min = $min;
function $max(values) { return _minmax(values, types_1.Ascending); }
exports.$max = $max;
function $average(values, opts = {}) {
    let [count, definedCount, okCount, sum] = _countsAndSum(values);
    if (opts.countsOnlyOKItems) {
        count = okCount;
    }
    else if (opts.countsOnlyDefinedItems) {
        count = definedCount;
    }
    return (0, commons_1.$defined)(sum) && count > 0 ? sum / count : undefined;
}
exports.$average = $average;
Array.prototype.average = function average(opts) { return $average(this, opts); };
Array.prototype.compare = function compare(other) { return (0, commons_1.$isarray)(other) ? (0, compare_1.$arraycompare)(this, other) : undefined; };
Array.prototype.filteredMap = function filteredMap(options) { return $map(this, options); };
Array.prototype.filteredSet = function filteredSet(callback) { return $mapset(this, callback); };
Array.prototype.includesequal = function includesequal(object) { return $includesequal(this, object); };
Array.prototype.includesvisual = function includesvisual(object) { return $includesvisual(this, object); };
Array.prototype.isEqual = function isEqual(other) { return (0, commons_1.$isarray)(other) && (0, compare_1.$arrayequal)(this, other); };
Array.prototype.first = function first() { return $first(this); };
Array.prototype.fusionEnumeration = function fusionEnumeration() { return this; };
Array.prototype.last = function first() { return $last(this); };
Array.prototype.max = function max() { return $max(this); };
Array.prototype.min = function min() { return $min(this); };
Array.prototype.sum = function sum() { return $sum(this); };
Array.prototype.toArray = function toArray() { return this; };
function _mapEquality(target, values, fn, includesFn) {
    const set = new Set();
    let index = 0;
    for (let v of values) {
        const mv = fn(v, index++);
        if ((0, commons_1.$ok)(mv) && !set.has(mv) && !includesFn(target, mv)) {
            target.push(mv);
            set.add(mv);
        }
    }
}
function _countsAndSum(values) {
    let sum = 0;
    let validCount = 0;
    let definedCount = 0;
    let totalCount = 0;
    if ((0, commons_1.$ok)(values)) {
        for (let v of values) {
            if ((0, commons_1.$defined)(v)) {
                if (v !== null) {
                    if ((0, commons_1.$defined)(sum)) {
                        let n = undefined;
                        if ((0, commons_1.$isnumber)(v)) {
                            n = v;
                        }
                        else if ((0, commons_1.$isstring)(v)) {
                            n = Number(v);
                        }
                        else if ((0, commons_1.$ismethod)(v, 'toNumber')) {
                            n = v.toNumber();
                        }
                        if (!(0, commons_1.$isnumber)(n)) {
                            sum = undefined;
                        }
                        else {
                            sum += n;
                        }
                    }
                    validCount++;
                }
                definedCount++;
            }
            totalCount++;
        }
    }
    return [totalCount, definedCount, validCount, sum];
}
function _minmax(values, compValue) {
    let ret = undefined;
    if ((0, commons_1.$ok)(values)) {
        for (let v of values) {
            if (!(0, commons_1.$ok)(v)) {
                return undefined;
            }
            else if (!(0, commons_1.$ok)(ret)) {
                ret = v;
            }
            else {
                let comp = (0, compare_1.$compare)(ret, v);
                if (!(0, commons_1.$defined)(comp)) {
                    return undefined;
                }
                else if (comp === compValue) {
                    ret = v;
                }
            }
        }
    }
    return ret;
}
//# sourceMappingURL=array.js.map